(* file: complementaryPalindromes.ml
 * author: Bob Muller
 *
 * CS1101 Computer Science I
 *
 * Find the start positions and lengths of complementary palindromes 
 * in fasta file sequences where the palindromes have lengths between
 * 4 and 12. This is the Location Restriction Sites problem from the
 * Rosalind bioinformatics site:
 *
 * http://rosalind.info/problems/revp/
 *)

(* readFasta : string -> string list
 *
 * The call (readFasta file) returns a list of strings making up the
 * lines in file.
 *)
let readFasta filename =
  let inch = open_in filename in
  let rec repeat lines =
    try
      repeat ((input_line inch)::lines)
    with
      End_of_file -> close_in inch;
                     lines
  in
  List.rev (repeat [])

(* complements : char -> char -> bool
 *)
let complements n1 n2 =
  match (n1, n2) with
  | ('A', 'T') | ('T', 'A') | ('C', 'G') | ('G', 'C') -> true
  | _ -> false

(* reversePalindrome : string -> bool
 *
 * The call (reversePalindrome seq) returns true of seq is a complementary
 * palindrome.
 *)
let reversePalindrome seq =
  let rec repeat i j =
    match i >= j with
    | true  -> true
    | false -> (complements seq.[i] seq.[j]) && repeat (i + 1) (j - 1)
  in
  repeat 0 (String.length seq - 1)

(* revps : string -> int -> int * int list
 *
 * This is a helper function for complementaryPalindromes (below). 
 * NB: eol abbreviates end-of-left, meaning index of rightmost 
 * character of left half of a complementary palindrome.
 *)
let revps seq eol =
  let rec repeat length acc =
    let start = eol - length / 2 in
    match start < 0 || (start + length) > String.length seq with
    | true  -> acc
    | false -> let str = String.sub seq start length
               in
               match reversePalindrome str with
               | true  -> repeat (length + 2) ((start, length) :: acc)
               | false -> acc
  in
  repeat 2 []

(* complementaryPalindromes : string -> int * int list
 *
 * The call (complementaryPalindromes seq) finds the starting positions 
 * and lengths of all complementary palindromic subsequences of seq.  
 * NB: Zero-based indexing rather than the 1-based indexing on Rosalind.
 *)
let complementaryPalindromes seq =
  let rec repeat eol acc =
    match eol >= String.length seq with
    | true  -> acc
    | false -> let pals = revps seq eol
               in
               repeat (eol + 1) (pals @ acc)
  in
  let matches = repeat 1 []
  in
  List.sort compare (List.filter (fun (_, len) -> len > 3) matches)

(* makeOutFile : string -> string
 *
 * The call (makeOutFile inFile) makes an output file name from an
 * input file name.
 *)
let makeOutFile name =
  let dot = String.index name '.'
  in
  (String.sub name 0 dot) ^ ".restrict"

(* writePairs : int * int list -> string -> unit
 *
 * The call (writePairs pairs outFile) writes out the starting positions
 * and lengths of all of the reverse palindromic sequences.
 *)
let writePairs pairs file =
  let ouch = open_out file
  in
  List.iter (fun (s, l) -> Printf.fprintf ouch "(%d, %d)\n" s l) pairs;
  close_out ouch

let _ =
  let inFile = Sys.argv.(1) in
  let outFile = makeOutFile inFile in
  let lines = readFasta inFile in
  let seq = List.hd (List.tl lines) in
  let pairs = complementaryPalindromes seq
  in
  writePairs pairs outFile

